# -*- coding: utf-8 -*-

import logging
from operator import itemgetter
from typing import List, Tuple, Callable, Dict

# TODO(davidr) is it faster to do this here or to save the function as an attribute of the
# class instance?
from timeit import default_timer

from collections import defaultdict

_logger = logging.getLogger(__name__)


class TimeMarker(object):
    """Time Marker class

    Todo:
        * how should this behave as a context manager vs. an instantiated object? right now, it
          only works as the former
        * we should support recurring tags (i.e. a timer.tag("foo") inside of a for loop and average
          the time taken for all the calls together, perhaps with stats?
    """

    __slots__ = [
        '_start', '_end', '_cumulative_time', '_last_tag', '_last_ts',
        '_timing'
    ]

    _reserved_tags = ['start', 'stop']
    _stats_formats = ['percentage', 'raw']

    def __init__(self):
        self._start: float = None
        self._end: float = None

        self._timing: Dict[str, float] = defaultdict(lambda: 0.0)

        self._cumulative_time: float = None

        self._last_tag: str = None
        self._last_ts: float = None

    def __enter__(self):
        self._start = default_timer()

        # Set up the initial values for our last_* tags
        self._last_ts = self._start
        self._last_tag = 'start'

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.tag("stop", _override=True)

        self._end = self._last_ts
        self._cumulative_time = self._end - self._start

    def __repr__(self) -> str:
        return "TimeMarker(start={}, last=({}, {}))".format(
            self._start, self._last_tag, self._last_ts)

    def tag(self, tag: str, _override: bool = False) -> None:
        """Mark a tagged event

        Args:
            tag: name of timer tag
            _override: override checking for reserved tags *Warning: don't use this!*

        Returns:
            None

        Raises:
            ValueError: if given invalid tag name

        """
        if tag in TimeMarker._reserved_tags and _override is False:
            raise ValueError('tag name {} is reserved'.format(tag))

        _ts = default_timer()
        self._cumulative_time = _ts - self._start

        self._timing[self._last_tag] += _ts - self._last_ts
        self._last_tag = tag
        self._last_ts = _ts

        _logger.debug("tag took {}".format(default_timer() - _ts))

    def stats(self,
              sort: bool = True,
              fmt: str = "percentage",
              oneline: bool = True,
              print_function: Callable = print,
              pctg_cap: float = 1.):
        """Print out TimeMarker stats

        Args:
            sort: if True, results are sorted by time taken, else in the order in which they were
                tagged
            fmt: format of stats. "raw" is time in seconds, "percentage" is the percentage of the
                total measured time
            oneline: if True, print stats on one line
            print_function: callable with string as only parameter
            pctg_cap: do not print results accounting for more than pctg_cap percentage of the total
                measured time

        Returns:
            None

        """
        if fmt not in ['percentage', 'raw']:
            raise ValueError(f'Invalid format: {fmt}')

        if not 0. < pctg_cap <= 1.:
            raise ValueError('pctg must be < 0., <= 1.')
        elif fmt not in TimeMarker._stats_formats:
            raise ValueError('"{:s}" not a valid stats formatter')
        elif pctg_cap < 1. and fmt != "percentage":
            raise ValueError('pctg_cap < 1. requires fmt="percentage"')
        elif pctg_cap < 1. and not sort:
            raise ValueError('pctg_cap < 1. requires sort=True')

        tag_separator = " " if oneline else "\n  - "
        tag_suffix = ""
        stats = list()
        timing_results = list(self._timing.items())

        stats.append("TIME:{:.6f}s".format(self._cumulative_time))

        if sort:
            # Sort by the time value of the tag
            timing_results = sorted(
                timing_results, key=itemgetter(1), reverse=True)

        if fmt == "raw":
            tag_suffix = "s"

        elif fmt == "percentage":
            timing_results = self._calculate_tag_percentage(
                self._cumulative_time, pctg_cap, timing_results)

        for tag, duration in timing_results:
            stats.append("{:s}:{:02.3f}{:s}".format(tag, duration, tag_suffix))

        stats_string = tag_separator.join(stats)
        print_function(stats_string)

    @staticmethod
    def _calculate_tag_percentage(cumulative_time, pctg_cap, timing_results):
        running_pctg = 0.
        tags_pctg: List[Tuple[str, float]] = list()
        for tag, ts in timing_results:
            tag_pctg = ts / cumulative_time

            # if we have a cap of the percentage of time we want accounted for, see if we've
            # gone over
            running_pctg += tag_pctg
            if running_pctg > pctg_cap:
                break

            tags_pctg.append((tag, tag_pctg))
        return tags_pctg



from setuptools import setup, find_packages

setup(
    name='timemarker',
    packages=find_packages(),
    version='0.2.0',
    author='David Ressman',
    author_email='davidr@ressman.org',
    license='MIT',
    description='Simple arbitrary time measurement tool',
    include_package_data=True,
    url='https://github.com/davidr/python-timemarker',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Utilities',
    ],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    extras_require=dict(
        test=['coveralls', 'pytest'],
        build=['sphinx', 'pkginfo', 'setuptools-git', 'twine', 'wheel', 'sphinxcontrib-napoleon']
    )
)





# -*- coding: utf-8 -*-

from time import sleep

import pytest

import timemarker

from .helpers import split_oneline


class TestBasics(object):

    def test_instantiate(self):
        timer = timemarker.TimeMarker()
        assert isinstance(timer, timemarker.TimeMarker)

    def test_contextmgr(self):
        with timemarker.TimeMarker() as timer:
            assert isinstance(timer, timemarker.TimeMarker)

        assert isinstance(timer, timemarker.TimeMarker)


class TestExceptions(object):

    def test_badtag(self):
        for tag_name in ['start', 'stop']:
            with pytest.raises(ValueError):
                with timemarker.TimeMarker() as timer:
                    timer.tag(tag_name)


class TestPctgTimer(object):

    def test_onetag_pctg(self, capsys):
        with timemarker.TimeMarker() as timer:
            timer.tag("sleep_100ms")
            sleep(.1)
        timer.stats()

        out, err = capsys.readouterr()
        assert err == ''
        assert out != ''
        out_dict = split_oneline(out)

        assert out_dict['TIME'] == pytest.approx(0.1, abs=1e-2)
        assert out_dict['start'] == pytest.approx(0.0, abs=1e-2)
        assert out_dict['sleep_100ms'] == pytest.approx(1.0, abs=1e-2)

    def test_twotag_pctg(self, capsys):
        with timemarker.TimeMarker() as timer:
            timer.tag("sleep_100ms_1")
            sleep(.1)
            timer.tag("sleep_100ms_2")
            sleep(.1)
        timer.stats()

        out, err = capsys.readouterr()
        assert err == ''
        assert out != ''
        out_dict = split_oneline(out)

        assert out_dict['TIME'] == pytest.approx(0.2, abs=1e-2)
        assert out_dict['start'] == pytest.approx(0.0, abs=1e-2)
        assert out_dict['sleep_100ms_1'] == pytest.approx(.5, abs=1e-2)
        assert out_dict['sleep_100ms_2'] == pytest.approx(.5, abs=1e-2)


class TestRawTimer(object):

    def test_onetag_raw(self, capsys):
        with timemarker.TimeMarker() as timer:
            timer.tag("sleep_100ms")
            sleep(.1)
        timer.stats(fmt="raw")

        out, err = capsys.readouterr()
        assert err == ''
        assert out != ''
        out_dict = split_oneline(out)

        assert out_dict['TIME'] == pytest.approx(0.1, abs=1e-2)
        assert out_dict['start'] == pytest.approx(0.0, abs=1e-2)
        assert out_dict['sleep_100ms'] == pytest.approx(.1, abs=1e-2)

    def test_twotag_raw(self, capsys):
        with timemarker.TimeMarker() as timer:
            timer.tag("sleep_100ms_1")
            sleep(.1)
            timer.tag("sleep_100ms_2")
            sleep(.1)
        timer.stats(fmt="raw")

        out, err = capsys.readouterr()
        assert err == ''
        assert out != ''
        out_dict = split_oneline(out)

        assert out_dict['TIME'] == pytest.approx(0.2, abs=1e-2)
        assert out_dict['start'] == pytest.approx(0.0, abs=1e-2)
        assert out_dict['sleep_100ms_1'] == pytest.approx(.1, abs=1e-2)
        assert out_dict['sleep_100ms_2'] == pytest.approx(.1, abs=1e-2)


class TestTagging(object):
    def test_tag_aggregation(self, capsys):
        with timemarker.TimeMarker() as timer:
            # Run this test twice and see if the the same tags are aggregated
            for i in range(2):
                timer.tag("sleep_100ms_1")
                sleep(.1)
                timer.tag("sleep_100ms_2")
                sleep(.1)
        timer.stats(fmt='raw')

        out, err = capsys.readouterr()
        assert err == ''
        assert out != ''
        out_dict = split_oneline(out)

        assert out_dict['TIME'] == pytest.approx(0.4, abs=1e-2)
        assert out_dict['start'] == pytest.approx(0.0, abs=1e-2)
        assert out_dict['sleep_100ms_1'] == pytest.approx(.2, abs=1e-2)
        assert out_dict['sleep_100ms_2'] == pytest.approx(.2, abs=1e-2)




from typing import List, Tuple, Dict
import re


def split_oneline(str: str) -> Dict[str, float]:
    ret_list = list()

    for tag in str.split(sep=' '):
        ret_list.append(_parse_tag(tag))

    return dict(ret_list)


def _parse_tag(tag: str) -> Tuple[str, float]:
    tag_name, tag_time = tag.split(sep=":")
    return tag_name, float(_filter_float(tag_time))


def _filter_float(my_string: str) -> str:
    return re.sub('[^0-9.]', "", my_string)
